#from csclib_tfhe import _csclib_tfhe as tfhe
#from . import pycsclib_tfhe

# csclib_tfhe._csclib_tfhe.PyLweInt
import csclib_tfhe._csclib_tfhe as tfhe
from . import pycsclib_tfhe
#from . import benes
#from . import LOUDS

LWE_TRUE = tfhe.PyLweShort(1)
LWE_FALSE = tfhe.PyLweShort(0)

print("csclib_tfhe imported")




#  An implementation of the control bits computation for Benes networks
#  based on Daniel J. Bernstein's paper
#@misc{cryptoeprint:2020/1493,
#      author = {Daniel J.  Bernstein},
#      title = {Verified fast formulas for control bits for permutation networks},
#      howpublished = {Cryptology {ePrint} Archive, Paper 2020/1493},
#      year = {2020},
#      url = {https://eprint.iacr.org/2020/1493}
#}



#python3 -m venv env
#source env/bin/activate
#maturin build --release
#pip install --force-reinstall ./rust/target/wheels/csclib_tfhe-*.whl 

#from csclib_tfhe import _csclib_tfhe as tfhe
from csclib_tfhe import *
from csclib_tfhe.benes import Benes_sort, AppPerm
from csclib_tfhe.pycsclib_tfhe import enc, enc_int, enc_array_int, enc_array, dec_array, blog

flatten = lambda x: [z for y in x for z in (flatten(y) if hasattr(y, '__iter__') and not isinstance(y, str) else (y,))]
flatten_dec = lambda x: [z.dec() for y in x for z in (flatten(y) if hasattr(y, '__iter__') and not isinstance(y, str) else (y,))]

def permutation(c):
    m = 1
    while (2*m-1)<<(m-1) < len(c): m += 1
    assert (2*m-1)<<(m-1) == len(c)
    n = 1<<m
    pi = list(range(n))
    pi = enc_array_int(pi, m)
    for i in range(2*m-1):
        gap = 1<<min(i,2*m-2-i)
        for j in range(n//2):
            #print("c[{}]: {}".format(i*n//2+j, c[i*n//2+j].dec()))
            pos = (j%gap)+2*gap*(j//gap)
            #if c[i*n//2+j] == encrypt_int(1, m, 1):
            #    pi[pos],pi[pos+gap] = pi[pos+gap],pi[pos]
            d = (c[i*n//2+j] == 1)
            pi[pos],pi[pos+gap] = d.cmux_int(pi[pos], pi[pos+gap])
    return pi

def composeinv(c,pi):
    #return [y for x,y in sorted(zip(pi,c))]
    print("composeinv input c:", flatten_dec(c))
    print("composeinv input pi:", flatten_dec(pi))
    n = len(pi)
    w = blog(n-1)+1
    sigma1, c1 = Benes_sort(c, w)
    pi2 = AppPerm(pi, sigma1, True)
    sigma2, pi3 = Benes_sort(pi2, w)
    y = AppPerm(c1, sigma2, True)
    print("y:", flatten_dec(y))
    return y


def lwe_min(x, y):
    if type(x) == int:
        x = enc_int(x, y.get_bit_size())
    elif type(y) == int:
        y = enc_int(y, x.get_bit_size())
    c = x < y
    return c.cmux_int(x, y)[0]

#def controlbits(pi):
def Perm_to_Benes(pi):
    n = len(pi)
    m = 1
    while 1<<m < n: m += 1
    assert 1<<m == n
    if m == 1: return [[pi[0]]]

    p = [pi[x^1] for x in range(n)]
    q = [pi[x]^1 for x in range(n)]
    range_n = [enc_int(x, m) for x in range(n)]
    piinv = composeinv(range_n,pi)
    p,q = composeinv(p,q),composeinv(q,p)

    #c = [lwe_min(encrypt_int(x, m), p[x]) for x in range(n)]
    c = [lwe_min(p[x], x) for x in range(n)]
    p,q = composeinv(p,q),composeinv(q,p)
    for i in range(1,m-1):
        cp,p,q = composeinv(c,q),composeinv(p,q),composeinv(q,p)
        c = [lwe_min(c[x],cp[x]) for x in range(n)]

    print("c:", [x.dec() for x in flatten(c)])
    f = [c[2*j] % 2 for j in range(n//2)]
    #F = [encrypt_int(x, m)^f[x//2] for x in range(n)]
    F = [x ^ f[x//2] for x in range(n)]
    Fpi = composeinv(F,piinv)
    l = [Fpi[2*k] % 2 for k in range(n//2)]
    #L = [encrypt_int(y, m)^l[y//2] for y in range(n)]
    L = [y ^ l[y//2] for y in range(n)]
    M = composeinv(Fpi,L)
    subM = [[M[2*j+e]>>1 for j in range(n//2)] for e in range(2)]
    print("subM:", flatten_dec(subM))
    #subz = map(Perm_to_Benes,subM)
    subz = [Perm_to_Benes(x) for x in subM]
    sz = []
    for i in range(len(subz[0])):
        sz.append([subz[j][i] for j in range(len(subz))])
    #z = [s for s0s1 in zip(*subz) for s in s0s1]
    #for s0s1 in zip(*subz):
    z = []
    for s0s1 in sz:
        #print("s0s1:", flatten_dec(s0s1))
        for s in s0s1:
            #print("s:", flatten_dec(s))
            z.append(s)
    print("z:", flatten_dec(z))
    return [f,z,l]

if __name__ == '__main__':
    pi = [3,2,1,0]
    n = len(pi)
    m = blog(n-1)+1
    pi_enc = enc_array_int(pi, m)
    c = Perm_to_Benes(pi_enc)
    print("c:", [x.dec() for x in flatten(c)])
    c = flatten(c)
    pi2 = permutation(c)
    pi2 = dec_array(pi2)
    print(pi2)
    assert pi2 == pi



#python3 -m venv env
#source env/bin/activate
#maturin build --release
#pip install --force-reinstall ./rust/target/wheels/csclib_tfhe-*.whl 


from csclib_tfhe import *
from csclib_tfhe.benes import Benes, Benes_apply, Benes_to_Perm, Perm_to_Benes, Perm_to_Benes_1bit, AppPerm, ApplySwitch, encrypt, encrypt_int, blog

################################################################
### X の i 番目の 0 と i 番目の 1 を交換する置換を求める．
### 0 と 1 の数は同じとする．
### O(n log^2 n) 時間
################################################################
def zero_one_match(X):
  n = len(X)
  w = blog(n-1)+1
  N, Y = Benes(X, half=True)
  I = [tfhe.PyLweInt(i, w) for i in range(n)]
  I2 = Benes_apply(I, N, False)
  for i in range(n//2):
    tmp = I2[i]
    I2[i] = I2[n-1-i]
    I2[n-1-i] = tmp
  I3 = Benes_apply(I2, N, True)
  return I3

################################################################
### LOUDS 表現に関する演算
### 次数 d のノードは 1 0^d と表現される
### 先頭にダミーの 0 がついているとする (super root の次数)
### super root の 1 は省略されているとする
### 長さは 2n ビット (n はノード数)
################################################################



################################################################
### LOUDS表現された木構造 t とノードのラベル v から
### 各ノードの親ノードのラベルを返す
### t: LOUDS表現された木構造
### v: 各ノードのラベル（整数）
### z: 各ノードの親ノードのラベル
### 根ノードの id は 0 とする．全体で n ノード．t の長さは 2n ビット
### O(n log^2 n) 時間
################################################################
def LOUDS_parentlabel(t, v, r, sigma=None):
  """LOUDS表現された木構造 t とノードのラベル v から各ノードの親ノードのラベルを返す
     根ノードの id は 0 とする．全体で n ノード．t の長さは 2n ビット
     O(n log^2 n) 時間

  Args:
      t (0,1列): LOUDS表現
      v (整数配列): 各ノードの値
      r (整数): super rootの値
      sigma (整数配列, optional): t の配列表現

  Returns:
      z (整数配列): 各ノードの親の値
  >>> X = enc_array([0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1])
  >>> V = enc_array_int([0,1,2,3,4,5,6,7,8,9], 4)
  >>> P = LOUDS_parentlabel(X, V, enc_int(7, 4))
  >>> dec_array(P)
  [7, 0, 0, 0, 1, 1, 1, 3, 5, 5]
  """
  n = (len(t)+1)//2 # ノード数
  m = len(t)

  if sigma is None:
    N, _ = Benes(t)
    sigma = [N]

  k = v[0].get_bit_size()
  d = [encrypt_int(0, k)] * n + [v[0] - r] + [v[i+1] - v[i] for i in range(n-1)]
#  d = tfhe.PyLweIntArray(d)
  w = AppPerm(d, sigma, inverse=True)
  x = [None] * m
#  x = tfhe.PyLweIntArray([encrypt_int(0, k)] * m)
  x[0] = r
  for i in range(1, m):
    x[i] = x[i-1] + w[i]
  y = AppPerm(x, sigma)
  z = y[0:n]

  return z

################################################################
### O(n log^2 n) 時間
################################################################
def LOUDS_firstchildlabel(t, v, sigma=None):
  """LOUDS表現 t とノードの値 v から，各ノードの最初の子ノードの値を返す

  Args:
      t (0,1列): LOUDS表現
      v (整数配列): 各ノードの値
      sigma (整数配列, optional): t の配列表現

  Returns:
      z (整数配列): 各ノードの最初の子ノードの値
  >>> X = enc_array([0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1])
  >>> V = enc_array_int([0,1,2,3,4,5,6,7,8,9], 4)
  >>> P = LOUDS_firstchildlabel(X, V)
  >>> dec_array(P)
  [1, 4, 7, 7, 8, 8, 9, 9, 9, 9]
  """
  n = (len(t)+1)//2 # ノード数
  m = len(t)

  if sigma is None:
    N, _ = Benes(t)
    sigma = [N]

  k = v[0].get_bit_size()
  #d =  [v[i+1] - v[i] for i in range(n-1)] + [encrypt_int(0, k) - v[n-1]] + [encrypt_int(0, k)] * n
  d =  [v[i+1] - v[i] for i in range(n-1)] + [encrypt_int(0, k)] + [encrypt_int(0, k)] * n
#  d = tfhe.PyLweIntArray(d)
  w = AppPerm(d, sigma, inverse=True)
  x = [None] * m
#  x = tfhe.PyLweIntArray([encrypt_int(0, k)] * m)
  x[0] = v[0] + w[0]
  for i in range(1, m):
    x[i] = x[i-1] + w[i]
  y = AppPerm(x, sigma)
  z = [None] * n
#  z = tfhe.PyLweIntArray([encrypt_int(0, k)] * n)
  #for i in range(0, n):
  #  z[i] = y[i+n]
  z = y[n:m]

  return z



################################################################
### 親と最初の子ノードのインデックスを返す
### (unary表現をbinary表現に変換しているだけ)
### parent(0) = -1 とする
### O(n log^2 n) 時間
################################################################
def LOUDS_parent_firstchild(t, sigma=None):
  """親と最初の子ノードのインデックスを返す

  Args:
      t (0,1列): LOUDS表現
      sigma (整数配列, optional): t の配列表現
  Returns:
      p (整数配列): 各ノードの親の配列表現
      fc (整数配列): 各ノードの最初の子の配列表現
  >>> X = enc_array([0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1])
  >>> V = enc_array_int([0,1,2,3,4,5,6,7,8,9], 4)
  >>> p, q = LOUDS_parent_firstchild(X)
  >>> dec_array(p)
  [63, 0, 0, 0, 1, 1, 1, 3, 5, 5]
  >>> dec_array(q)
  [1, 4, 7, 7, 8, 8, 10, 10, 10, 10]
  """
  n = (len(t)+1)//2 # ノード数
  m = len(t)
  k = blog(m-1)+1

  if sigma is None:
    N, _ = Benes(t)
    sigma = [N]

#  id = [tfhe.PyLweInt(i, k) for i in range(m)]
#  sigma_inv = AppPerm(id, sigma)
  sigma_inv = Benes_to_Perm(sigma)

### firstchild(i) = rank0( select1(i+1) ) = select1(i+1) - i (i >= 0)
### parent(i) = rank1( select0(i+1) ) - 1 = select0(i+1) - i - 1 (i >= 0)
### sigma_inv = [select0(i+1)] + [select1(i+1)]
  p = [None] * n
#  p = tfhe.PyLweIntArray([encrypt_int(0, k)] * n)
  fc = [None] * n
#  fc = tfhe.PyLweIntArray([encrypt_int(0, k)] * n)
  for i in range(0, n):
    p[i] = sigma_inv[i] - (i + 1)
    fc[i] = sigma_inv[i+n] - i

  return p, fc


################################################################
### LOUDS表現 t から，縮約（親の親を新しい親にする）した木のLOUDS表現 t2 と
### その配列表現 rho を返す
### sigma は t の配列表現
### O(n log^2 n) 時間
################################################################
def LOUDS_contract(t, sigma=None):
  """LOUDS表現 t から縮約 (親の親を新しい親にする) した木のLOUDS表現 t2 とその配列表現 rho を返す
  Args:
      t (0,1列): LOUDS表現
      sigma (整数配列, optional): t の配列表現

  Returns:
      t2 (0,1列): LOUDS表現
      rho (整数配列): t2 の配列表現
  >>> X = enc_array([0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1])
  >>> t2, rho = LOUDS_contract(X)
  >>> dec_array(t2)
  [0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1]
  >>> dec_array(rho)
  [0, 1, 2, 3, 5, 6, 7, 8, 10, 11, 4, 9, 12, 13, 14, 15, 16, 17, 18, 19]
  """
  if sigma is None:
    N, _ = Benes(t)
    sigma = [N]

  n = len(t)
  p, fc = LOUDS_parent_firstchild(t, sigma)
#  print("p:", [pi.dec() for pi in p])
#  print("fc:", [qi.dec() for qi in fc])

  #p2 = LOUDS_parentlabel(t, p, encrypt_int(0, p[0].get_bit_size()))
  p2 = LOUDS_parentlabel(t, p, p[0], sigma)
  fc2 = LOUDS_firstchildlabel(t, fc, sigma)
#  print("p2:", [pi.dec() for pi in p2])
#  print("fc2:", [qi.dec() for qi in fc2])
  rho = [None] * n
#  rho = tfhe.PyLweIntArray([tfhe.PyLweInt(0, p2[0].get_bit_size())] * n)
  for i in range(n//2):
    rho[i] = p2[i] + (i+1)
    rho[i + n//2] = fc2[i] + i
#  print("rho:", [ri.dec() for ri in rho])

#  tmp = [encrypt(0)] * (n//2) + [encrypt(1)] * (n//2)
#  t2 = AppPerm(tmp, Perm_to_Benes(rho))
#  return t2, rho

### rho を拡張して長さを2冪にする
  k = blog(n-1)+1
  m = 1 << k
  diff = (m - n)//2
  t2tmp = [encrypt(0)] * (m//2) + [encrypt(1)] * (m//2)
  if diff > 0:
    rhotmp = [encrypt_int(i, k) for i in range(0, diff)] + [x+diff for x in rho] + [encrypt_int(i, k) for i in range(n+diff, m)]
    t2 = AppPerm(t2tmp, [Perm_to_Benes_1bit(rhotmp)])
    t2 = t2[diff : diff + n]
  else:
    t2 = AppPerm(t2tmp, [Perm_to_Benes_1bit(rho)])

  return t2, rho

################################################################
### LOUDS表現 t とノードの値 v から，各ノードの子の値の和を返す
### sigma は t の配列表現
### O(n log^2 n) 時間
################################################################
def LOUDS_childlabelSum(t, v, sigma=None):
  """LOUDS表現 t とノードの値 v から，各ノードの子の値の和を返す

  Args:
      t (0,1列): LOUDS表現
      v (整数配列): 各ノードの値
      sigma (整数配列, optional): t の配列表現

  Returns:
      z (整数配列): 各ノードの子の値の和
  >>> X = enc_array([0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1])
  >>> v = enc_array_int([1,1,1,1,1,1,1,1,1,1], 4)
  >>> z = LOUDS_childlabelSum(X, v)
  >>> dec_array(z)
  [3, 3, 0, 1, 0, 2, 0, 0, 0, 0]
  """
  n = (len(t)+1)//2 # ノード数
  m = len(t)

  if sigma is None:
    N, _ = Benes(t)
    sigma = [N]

  k = v[0].get_bit_size()
  vp = v + [encrypt_int(0, k)] * n
  #vp = v @ ([encrypt_int(0, k)] * n)
  w = AppPerm(vp, sigma, inverse=True)
  x = [None] * m
#  x = tfhe.PyLweIntArray([encrypt_int(0, k)] * m)
  x[0] = w[0]
  for i in range(1, m):
    x[i] = x[i-1] + w[i]
  y = AppPerm(x, sigma)
  z = [None] * n
#  z = tfhe.PyLweIntArray([encrypt_int(0, k)] * n)
  for i in range(0, n-1):
    z[i] = y[i+1+n] - y[i+n]
  z[n-1] = v[0]
  for i in range(1, n):
    z[n-1] = z[n-1] + v[i]
  z[n-1] = z[n-1] - y[2*n-1]

  return z


flatten = lambda x: [z for y in x for z in (flatten(y) if hasattr(y, '__iter__') and not isinstance(y, str) else (y,))]
flatten_dec = lambda x: [z.dec() for y in x for z in (flatten(y) if hasattr(y, '__iter__') and not isinstance(y, str) else (y,))]

def LOUDS_contract_all(t):
  """LOUDS表現 t を繰り返し縮約していき，その過程で得られる LOUDS 表現と配列表現を返す
  Args:
      t (0,1列): LOUDS表現
  Returns:
      T (0,1列の配列): 縮約過程で得られる LOUDS 表現
      Sigma (整数配列の配列): 縮約過程で得られる配列表現
  >>> X = enc_array([0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1])
  >>> T_Sigma = LOUDS_contract_all(X)
  >>> z = LOUDS_PathSum(X_enc, v, T_Sigma)
  >>> dec_array(z)
  [1, 2, 2, 2, 3, 3, 3, 3, 4, 4]
  """
  n = len(t) // 2
  k = blog(n-1)+1

  T = [None] * k
  Sigma = [None] * k

  for r in range(0, k):
    N, _ = Benes(t)
    sigma = [N]
    T[r] = t
    Sigma[r] = sigma
    if r < k-1:    
      print("contract_all r:", r, flatten_dec(t))
      t, sigma_int = LOUDS_contract(t, sigma)
  return T, Sigma

################################################################
### LOUDS表現 t とノードの値 v から，根から各ノードへのパス上の値の和を返す
### sigma は t の配列表現
### O(n log^3 n) 時間
################################################################
def LOUDS_PathSum(t_, v, t_sigma=None):
  """LOUDS表現 t とノードの値 v から，根から各ノードへのパス上の値の和を返す
     O(n log^3 n) 時間

  Args:
      t (0,1列): LOUDS表現
      v (整数配列): 各ノードの値
      t_sigma (整数配列, 整数配列, optional): t と sigma を事前計算したもの

  Returns:
      z (整数配列): 各ノードの子の値の和
  >>> X = enc_array([0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1])
  >>> v = enc_array_int([1,1,1,1,1,1,1,1,1,1], 4)
  >>> z = LOUDS_PathSum(X, v)
  >>> dec_array(z)
  [1, 2, 2, 2, 3, 3, 3, 3, 4, 4]
  """
#  if sigma is None:
#    N, _ = Benes(t)
#    sigma = [N]
  t = t_.copy()

  n = len(t) // 2
  k = blog(n-1)+1
  z = v.copy()
  #z = v.dup()
  for r in range(0, k):
    #p = LOUDS_parentlabel(t, z, encrypt_int(0, v[0].get_bit_size()), sigma)
    if t_sigma is not None:
      t = t_sigma[0][r]
      sigma = t_sigma[1][r]
      print("PathSum r:", r, flatten_dec(t), flatten_dec(z))
      p = LOUDS_parentlabel(t, z, encrypt_int(0, v[0].get_bit_size()), sigma)
    else:
      N, _ = Benes(t)
      sigma = [N]
      print("PathSum r:", r, flatten_dec(t), flatten_dec(z))
      p = LOUDS_parentlabel(t, z, encrypt_int(0, v[0].get_bit_size()), sigma)
      t, sigma_int = LOUDS_contract(t, sigma)
    for i in range(n):
      z[i] = z[i] + p[i]
    #z = z + p
  return z

################################################################
### LOUDS表現 t とノードの値 v から，各ノードの子孫の値の和を返す
### sigma は t の配列表現
### O(n log^3 n) 時間
################################################################
def LOUDS_TreeSum(t_, v, t_sigma=None):
  """LOUDS表現 t とノードの値 v から，各ノードの子孫の値の和を返す
     sigma は t の配列表現
     O(n log^3 n) 時間
  Args:
      t (0,1列): LOUDS表現
      v (整数配列): 各ノードの値
      t_sigma (整数配列, 整数配列, optional): t と sigma を事前計算したもの

  Returns
      z (整数配列): 各ノードの子孫の値の和
  >>> X = enc_array([0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1])
  >>> v = enc_array_int([1,1,1,1,1,1,1,1,1,1], 4)
  >>> z = LOUDS_TreeSum(X, v)
  >>> dec_array(z)
  [10, 6, 1, 2, 1, 3, 1, 1, 1, 1]
  """
  t = t_.copy()

  n = len(t) // 2
  k = blog(n-1)+1
  z = v.copy()
  for r in range(0, k):
    #print("TreeSum r:", r, flatten_dec(t), flatten_dec(z))
    if t_sigma is not None:
      t = t_sigma[0][r]
      sigma = t_sigma[1][r]
      #print("TreeSum r:", r, flatten_dec(t), flatten_dec(z))
      s = LOUDS_childlabelSum(t, z, sigma)
    else:
      N, _ = Benes(t)
      sigma = [N]
      #print("TreeSum r:", r, flatten_dec(t), flatten_dec(z))
      s = LOUDS_childlabelSum(t, z, sigma)
      t, sigma_int = LOUDS_contract(t, sigma)
    for i in range(n):
      z[i] = z[i] + s[i]
    #z = z + s
  return z

################################################################
### LOUDS表現 t から，各ノードの深さを返す
### 根の深さは 1 とする
### sigma は t の配列表現
### O(n log^3 n) 時間
################################################################
def LOUDS_depth(t, t_sigma=None):
  """LOUDS表現 t から，各ノードの深さを返す
     根の深さは 1 とする
     sigma は t の配列表現
     O(n log^3 n) 時間
  >>> X = enc_array([0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1])
  >>> z = LOUDS_depth(X_enc)
  >>> dec_array(z)
  [1, 2, 2, 2, 3, 3, 3, 3, 4, 4]
  """
  n = (len(t)+1)//2 # ノード数
  k = blog(n+1-1)+1 # 値の範囲は [0,n]
  depth = [encrypt_int(1, k)] * n

  return LOUDS_PathSum(t, depth, t_sigma)

################################################################
def LOUDS_desc(t, t_sigma=None):
  """LOUDS表現 t とノードの値 v から，各ノードの子孫の数を返す
     sigma は t の配列表現
     O(n log^3 n) 時間
  >>> X = enc_array([0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1])
  >>> z = LOUDS_desc(X_enc)
  >>> dec_array(z)
  [10, 6, 1, 2, 1, 3, 1, 1, 1, 1]
  """
  n = (len(t)+1)//2 # ノード数
  k = blog(n+1-1)+1 # 値の範囲は [0,n]
  desc = [encrypt_int(1, k)] * n

  return LOUDS_TreeSum(t, desc, t_sigma)


if __name__ == '__main__':
  import doctest
  from csclib_tfhe.pycsclib_tfhe import enc_int, enc_array_int, enc_array, dec_array
  #doctest.testmod()

#  X = [0,1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1]
#  #X = [1,0,0,0,1,0,0,0,1,1,0,1,1,0,0,1,1,1,1]
  X = [0,1,0,0,1,0,1,1]
  n = len(X)
  k = blog(10)+1
  X_enc = [encrypt(x) for x in X]
#  X_enc = tfhe.PyLweShortArray([encrypt(x) for x in X])

#
#  V_enc = [tfhe.PyLweInt(i, k) for i in range((n+1)//2)]

#  P = LOUDS_parentlabel(X_enc, V_enc, encrypt_int(7, k))
#  print("P:", [pi.dec() for pi in P])

#  p, q = LOUDS_parent_firstchild(X_enc)
#  print("p:", [pi.dec() for pi in p])
#  print("q:", [qi.dec() for qi in q])

#  P = LOUDS_firstchildlabel(X_enc, V_enc)
#  print("P:", [pi.dec() for pi in P])

#  t2, rho = LOUDS_contract(X_enc)
#  print("rho:", [ri.dec() for ri in rho])
#  print("t2:", [ti.dec() for ti in t2])
#  print("X_enc:", [xi.dec() for xi in X_enc])

  v = [encrypt_int(1, k) for i in range(n//2)]
#  v = tfhe.PyLweIntArray(v)

  t_sigma = LOUDS_contract_all(X_enc)

  z = LOUDS_PathSum(X_enc, v, t_sigma)
  print("z:", [zi.dec() for zi in z])
#  print("X_enc:", [xi.dec() for xi in X_enc])
#  print("v:", [zi.dec() for zi in v])
#  z2 = LOUDS_childlabelSum(X_enc, v)
#  print("z2:", [zi.dec() for zi in z2])
#  z3 = LOUDS_TreeSum(X_enc, v)
#  print("z3:", [zi.dec() for zi in z3])



#python3 -m venv env
#source env/bin/activate
#maturin build --release
#pip install --force-reinstall ./rust/target/wheels/csclib_tfhe-*.whl 


from csclib_tfhe import * # ok
from csclib_tfhe.benes import ApplySwitch, butterfly, inv_butterfly, Benes_apply, AppPerm
from csclib_tfhe.pycsclib_tfhe import enc, enc_int, enc_array_int, enc_array, dec_array, blog
#import random

#def encrypt(x):
#  return tfhe.PyLweShort(x)

#def encrypt_int(x, bit):
#  return tfhe.PyLweInt(x, bit)

#def enc_array(arr, bit):
#  if bit == 0:
#    return [tfhe.PyLweShort(val) for val in arr]
#  else:
#    return [tfhe.PyLweInt(val, bit) for val in arr]

#def dec_array(arr):
#  return [val.dec() for val in arr]

#def blog(x):
#  l = -1
#  while x > 0:
#    x >>= 1
#    l += 1
#  return l

###################################################################
### GenCycleを表すBenesネットワークを構成
### 1 を右ローテイト（左ローテイトは作ったネットワークの inverse を使う）
### 0 - 0       0 - 0
### 0 - 0       0 - 0
###
### 1 - 1       1   1
###               x
### 1 - 1       1   1
###
### 0   1       0   1
###   x           x
### 1   0       1   0
###
### 1 - 1       1 - 1
### 0 - 0       0 - 0
###################################################################
def GenCycle_Benes_sub(X):
  n = len(X)
  if n == 2:
    Y = [None] * n
    c = X[0] & X[1]
    #print("X:", X)
    #print("c:", c)
    Y[0], Y[1] = ApplySwitch(c, X[0], X[1])
    #print("Y:", Y)
    return ([[c],[],[],[]], Y)

  SL = [None] * (n//2)
  SR = [None] * (n//2)
  i = 0
  while i < n//2:
    SL[i] = (~X[i*2]) & X[i*2+1] # pythonの ~ はビット反転なので注意
    SR[i] = X[i*2+1]
    i += 1

  W = [0] * n
  i = 0
  while i < n//2:
    W[i*2], W[i*2+1] = ApplySwitch(SL[i], X[i*2], X[i*2+1])
    i += 1
  Wp = butterfly(W)

  XU = Wp[0:n//2]
  XL = Wp[n//2:n]

  netU, ZU = GenCycle_Benes_sub(XU)
  netL = []
  ZL = XL

  Zp = ZU + ZL
  Z = inv_butterfly(Zp)

  Y = [None] * n
  i = 0
  while i < n//2:
    c = SR[i]
    Y[i*2], Y[i*2+1] = ApplySwitch(c, Z[i*2], Z[i*2+1])
    i += 1
  return ([SL, netU, netL, SR], Y)

def GenCycle_Benes(X):
  """X の GenCycle を表す Benes ネットワークを返す

  Args:
      X (0,1列): ビット列

  Returns:
      N (Benes): X の GenCycle を表す Benes ネットワーク
      Y (0,1列): ビット列 (不要?)
  """
  n = len(X)
  d = blog(n-1)+1
  m = 1 << d
  if n < m:
    X_ext = X + [enc(1)] * (m - n)
    N, Y_ext = GenCycle_Benes_sub(X_ext)
    Y = Y_ext[0:n]
  else:
    N, Y = GenCycle_Benes_sub(X)
  return N, Y


def PrefixSum(v):
  """整数配列の接頭辞和を計算 (その場所を含む)

  Args:
      v (整数配列): 整数の配列

  Returns:
      z (整数配列): 接頭辞和配列 (その場所を含む)
  """
  n = len(v)
  z = [None] * n
  s = enc_int(0, v[0].get_bit_size())
  for i in range(n):
    s = s + v[i]
    z[i] = s
  return z

###################################################################
### Propagate
###################################################################
def Propagate(g, v):
  """各グループの先頭の要素をコピーする

  Args:
      g (0,1列): グループベクトル
      v (整数配列): 整数配列

  Returns:
      v (整数配列): 整数配列
  """
  bit = v[0].get_bit_size()
  pi, _ = GenCycle_Benes(g)
  #print("v:", dec_array(v, bit))
  x = AppPerm(v, [pi], True)
  #print("x:", x)
  #x[0] = encrypt_int(0, bit)
  x[0] = enc_int(0, bit)
  #print("v:", dec_array(v, bit))
  #print("x:", dec_array(x, bit))
  v2 = [v[i] - x[i] for i in range(len(v))]
  z = PrefixSum(v2)
  return z


if __name__ == '__main__':

    X = [1, 0, 1, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1]
    #X = [1, 0]
    #X = [random.randint(0, 1) for _ in range(16)]
    print("X:", X)
    X_enc = enc_array(X)
    net, Y_enc = GenCycle_Benes(X_enc)

    n = len(X)
    w = blog(n-1)+1

    A = [i for i in range(n)]
    A_enc = enc_array_int(A, w)
    print("A_enc", dec_array(A_enc))

    B = Benes_apply(A_enc, net)
    B_dec = dec_array(B)
    #print(B_dec)
    C = Benes_apply(A_enc, net, inverse=True)
    C_dec = dec_array(C)
    #print(C_dec)

    #print("Original X:", X)
    #print("Original A:", A)
    print("After Benes B:", B_dec)
    print("After Benes C:", C_dec)

    p = Propagate(X_enc, A_enc)
    p_dec = dec_array(p)
    print("After Propagate p:", p_dec)



#python3 -m venv env
#source env/bin/activate
#maturin build --release
#pip install --force-reinstall ./rust/target/wheels/csclib_tfhe-*.whl 


from csclib_tfhe import * # ok
from csclib_tfhe.benes import Benes, Benes_apply


###############################################################
### 各位置で，それより左の接頭辞での or を計算
###############################################################
def AllPrefixOr(X):
  """各位置で，それより左の接尾辞での or を計算 (その場所を含む)

  Args:
      X (0,1列): bit列

  Returns:
      (0,1列): 各位置での接尾辞 or
  """
  n = len(X)
  O = [None] * n
  r0 = LWE_FALSE
  for i in range(n):
    r0 = r0 | X[i]
    O[i] = r0
  return O

###############################################################
### 各位置で，それより右の接尾辞での or を計算
### （先頭は必ず 0 になる）
###############################################################
def AllSuffixOr(X, inclusive=False):
  """各位置で，それより右の接尾辞での or を計算

  Args:
      X (0,1列): bit列
      inclusive (bool, optional): True の場合，その場所を含んだ or を計算.  

  Returns:
      (0,1列): 各位置での接尾辞 or
  """
  n = len(X)
  O = [None] * n
  r = LWE_FALSE
  for i in range(n-1, -1, -1):
    if inclusive:
      r = r | X[i]
    O[i] = r
    if not inclusive:
      r = r | X[i]
  return O


################################################################
### w ビット整数 x の unit vector 表現を返す
### O(n) 時間
################################################################
def unitv(x):
    """w ビット整数 x の unit vector 表現を返す
       O(n) 時間

    Args:
        x (整数): w ビット整数

    Returns:
        F (0,1列): x の unit vector 表現 (F[x] = 1, その他 0)
    >>> dec_array(unitv(tfhe.PyLweInt(3, 3)))
    [0, 0, 0, 1, 0, 0, 0, 0]
    >>> dec_array(unitv(tfhe.PyLweInt(5, 3)))
    [0, 0, 0, 0, 0, 1, 0, 0]
    """
    w = x.get_bit_size()
    n = 1 << w

    x_bits = x.to_bit()

    F = [LWE_TRUE]
    for d in range(w):
        print("d:", d)
        s = 1 << d
        F_next = [None] * (s*2)
        b = x_bits[w-1-d]
        for i in range(s):
            F_next[2*i] = F[i] & ~b
            F_next[2*i+1] = F[i] & b
        F = F_next
    return F

################################################################
### w bit配列 V を整数 k だけ左回転
### O(wn log n) 時間
################################################################
def rotate_left(V, k):
    """w bit配列 V を整数 k だけ左回転
       O(wn log n) 時間

    Args:
        V (整数配列): w bit整数の配列
        k (整数): 回転量

    Returns:
        V_rot (整数配列): 左回転後の配列
    >>> dec_array(rotate_left(tfhe.PyLweInt.array([0,1,2,3,4,5,6,7], 3), tfhe.PyLweInt(2,3)))
    [2, 3, 4, 5, 6, 7, 0, 1]
    >>> dec_array(rotate_left(tfhe.PyLweInt.array([0,1,2,3,4,5,6,7], 3), tfhe.PyLweInt(5,3)))
    [5, 6, 7, 0, 1, 2, 3, 4]
    """
    n = len(V)
    u = unitv(k)
    o = AllSuffixOr(u)
    pi, _ = Benes(o)
    V_rot = Benes_apply(V, pi, False)
    return V_rot

################################################################
### w bit配列 V を整数 k だけ右回転
### O(wn log n) 時間
################################################################
def rotate_right(V, k):
    """w bit配列 V を整数 k だけ右回転
       O(wn log n) 時間

    Args:
        V (整数配列): w bit整数の配列
        k (整数): 回転量

    Returns:
        V_rot (整数配列): 右回転後の配列
    """
    n = len(V)
    u = unitv(k)
    o = AllSuffixOr(u)
    pi, _ = Benes(o)
    V_rot = Benes_apply(V, pi, True)
    return V_rot

def dec_array(arr):
  return [val.dec() for val in arr]

if __name__ == '__main__':
#
#    w = 3
#
#    A = [tfhe.PyLweInt(i, w) for i in range(1<<w)]
#
#    for x in range(1<<w):
#        print("x:", x)
#        y = rotate_right(A, tfhe.PyLweInt(x, w))
#        print("rol:", [val.dec() for val in y])
    import doctest
    doctest.testmod()



#from benes import Benes
from csclib_tfhe import _csclib_tfhe as tfhe
from csclib_tfhe import LWE_FALSE, LWE_TRUE

def blog(x):
  l = -1
  while x > 0:
    x >>= 1
    l += 1
  return l

def encrypt(x):
  return tfhe.PyLweShort(x)
def encrypt_int(x, bit):
  return tfhe.PyLweInt(x, bit)

###################################################################
### 平文の整数配列を暗号化
### 整数は bit ビットで表現される
### bit == 0 の場合はブール値の配列とみなす
###################################################################
def enc_array(arr, bit):
  if bit == 0:
    return [tfhe.PyLweShort(val) for val in arr]
  else:
    return [tfhe.PyLwePacked(val, bit) for val in arr]

#def ApplySwitch(c, x, y):
#  if type(x) == int:
#    x = encrypt(x)
#  elif type(y) == int:
#    y = encrypt(y)
#  xp, yp = c.cmux(x, y)
#  return xp, yp

#def ApplySwitch_int(c, x, y):
#  if type(x) == int:
#    x = encrypt_int(x, y.get_bit_size())
#  elif type(y) == int:
#    y = encrypt_int(y, x.get_bit_size())
#  xp, yp = c.cmux_int(x, y)
#  return xp, yp

def ApplySwitch(c, x, y):
  if type(x) == tfhe.PyLweShort:
    if type(y) == int:
      y = encrypt(y)
    xp, yp = c.cmux(x, y)
  elif type(y) == tfhe.PyLweShort:
    if type(x) == int:
      x = encrypt(x)
    xp, yp = c.cmux(x, y)
  elif type(x) == tfhe.PyLweInt:
    if type(y) == int:
      y = encrypt_int(y, x.get_bit_size())
    xp, yp = c.cmux_int(x, y)
  elif type(y) == tfhe.PyLweInt:
    if type(x) == int:
      x = encrypt_int(x, y.get_bit_size())
    xp, yp = c.cmux_int(x, y)
  elif type(x) == tfhe.PyLwePacked:
    if type(y) == int:
      y = tfhe.PyLwePacked(y, x.get_bit_size())
    xp, yp = c.cmux_packed(x, y)
  elif type(y) == tfhe.PyLwePacked:
    if type(x) == int:
      x = tfhe.PyLwePacked(x, y.get_bit_size())
    xp, yp = c.cmux_packed(x, y)
  else:
    print("ApplySwitch: unsupported type:", type(x), type(y))
    raise TypeError("ApplySwitch: unsupported type")

  return xp, yp

def butterfly(X):
  n = len(X)
  Y = [None] * n
  for i in range(0,n//2):
    Y[i] = X[2*i]
    Y[i+n//2] = X[2*i+1]
  return Y

def inv_butterfly(X):
  n = len(X)
  Y = [None] * n
  for i in range(0,n//2):
    Y[i*2] = X[i]
    Y[i*2+1] = X[i+n//2]
  return Y

def AllPrefixSumMod2(X):
  n = len(X)
  rank = [None] * n
  r0 = LWE_FALSE
  i = 0
  while i < n:
    r0 = r0 ^ (~X[i])
    rank[i] = r0
    i += 1
  return rank

def Benes_sub(X, half = False):
  n = len(X)
  print("Benes_sub n:", n)
  if n == 2:
    Y = [None] * n
#    Y = tfhe.PyLweShortArray([encrypt(0)] * n)
    c = X[0] & (~X[1])
    Y[0], Y[1] = ApplySwitch(c, X[0], X[1])
    return ([[c],[],[],[]], Y)

  rank = AllPrefixSumMod2(X)
  SL = [None] * (n//2)
#  SL = tfhe.PyLweShortArray([encrypt(0)] * (n//2))
  i = 0
  while i < n//2:
    SL[i] = ~rank[i*2]
    i += 1
  SR = ~rank[n-1]

  W = [None] * n
#  W = tfhe.PyLweShortArray([encrypt(0)] * n)
  i = 0
  while i < n//2:
    W[i*2], W[i*2+1] = ApplySwitch(SL[i], X[i*2], X[i*2+1])
    i += 1
  Wp = butterfly(W)

  XU = Wp[0:n//2]
  XL = Wp[n//2:n]

  netU, ZU = Benes_sub(XU, half)
  netL, ZL = Benes_sub(XL, half)
  Zp = ZU + ZL

  SR_new = []
  if half == False:
    Z = inv_butterfly(Zp)
    Y = [0] * n
#    Y = tfhe.PyLweShortArray([encrypt(0)] * n)
    i = 0
    while i < n//2:
      c = SR & Z[i*2]
      Y[i*2], Y[i*2+1] = ApplySwitch(c, Z[i*2], Z[i*2+1])
      SR_new.append(c)
      i += 1
#    SR_new = tfhe.PyLweShortArray(SR_new)
  else:
    Y = Zp
  return ([SL, netU, netL, SR_new], Y)

def Benes(X, half = False):
  """ビット列の安定ソートを表すBenesネットワークを返す

  Args:
      X (0,1列): ビット列
      half (bool, optional): True の場合は，ネットワークの前半だけ計算 (0,1 の数が等しい場合に使用)

  Returns:
      N (Benes): 安定ソートを表すBenesネットワーク
      Y (0,1列): ソート後のビット列
  """
  n = len(X)
  d = blog(n-1)+1
  m = 1 << d
  if n < m:
    X_ext = X + [encrypt(1)] * (m - n)
#    print("X type:", type(X), X)
#    X_ext = X @ ([encrypt(1)] * (m - n))
    N, Y_ext = Benes_sub(X_ext, half)
    Y = Y_ext[0:n]
  else:
    N, Y = Benes_sub(X, half)
  return N, Y

def Benes_apply_sub(x, network, inverse: bool):
  n = len(x)
  x_new = x.copy()
  sl_tmp, netU, netL, sr_tmp = network

#  print("Benes_apply_sub n:", n)
#  print("sl_tmp:", [v.dec() for v in sl_tmp])

  if n > 2:
    if inverse:
      sl = sr_tmp
      sr = sl_tmp
    else:
      sl = sl_tmp
      sr = sr_tmp
    i = 0
    if sl != []:
      while i < n:
        c = sl[i//2]
        x_new[i], x_new[i+1] = ApplySwitch(c, x_new[i], x_new[i+1])
        i += 2
      x_tmp = butterfly(x_new)
    else:
      x_tmp = x_new
    if netU != []:
      x_tmp1 = Benes_apply_sub(x_tmp[0:n//2], netU, inverse)
    else:
      x_tmp1 = x_tmp[0:n//2]
    if netL != []:
      x_tmp2 = Benes_apply_sub(x_tmp[n//2:n], netL, inverse)
    else:
      x_tmp2 = x_tmp[n//2:n]

    x_new = x_tmp1 + x_tmp2
    if sr != []:
      x_new = inv_butterfly(x_new)

      i = 0
      while i < n:
        c = sr[i//2]
        x_new[i], x_new[i+1] = ApplySwitch(c, x_new[i], x_new[i+1])
        i += 2

  else: # n == 2
    if sl_tmp != []:
      x_new[0], x_new[1] = ApplySwitch(sl_tmp[0], x[0], x[1])
    elif sr_tmp != []:
      x_new[0], x_new[1] = ApplySwitch(sr_tmp[0], x[0], x[1]) # Benesネットワークが正規化されていれば不要だが念のため入れておく
    else:
      x_new = x
  return x_new

###################################################################
### w-bit整数の配列 X にBenesネットワークを適用
### O(wn log n) time
###################################################################
def Benes_apply(X, network, inverse = False):
  """w-bit整数の配列 X にBenesネットワークを適用
     O(wn log n) time

  Args:
      X (整数配列): w-bit整数の配列 
      network ([Benes]): Benesネットワークの配列
      inverse (bool, optional): True の場合は逆置換を計算

  Returns:
      X_new (整数配列): w-bit整数の配列を network で置換した配列
  """
  n = len(X)
  d = blog(n-1)+1
  m = 1 << d
  if n < m:
    if type(X[0]) == tfhe.PyLweInt:
      #X = X @ ([encrypt_int(0, X[0].get_bit_size())] * (m - n))
      X = X + ([encrypt_int(0, X[0].get_bit_size())] * (m - n))
    elif type(X[0]) == tfhe.PyLweShort:
      X = X + [encrypt(0)] * (m - n)
    elif type(X[0]) == tfhe.PyLwePacked:
      #X = X @ ([tfhe.PyLwePacked(0, X[0].get_bit_size())] * (m - n))
      X = X + ([tfhe.PyLwePacked(0, X[0].get_bit_size())] * (m - n))

  X_new = Benes_apply_sub(X, network, inverse)
  X_new = X_new[0:n]
  return X_new

###################################################################
### ビットの配列で表された置換を 1 個のBenesネットワークに変換
###################################################################
def Bit_to_Benes(b):
  """ビットの配列で表された置換を 1 個のBenesネットワークに変換

  Args:
      b (0,1列): ビット列

  Returns:
      [Benes]: b の安定ソートを表す Benes ネットワークの配列(長さ 1)
  """
  net, _ = Benes(b)
  return [net]

###################################################################
### 整数の配列で表された置換を log n 個のBenesネットワークに変換
### O(n log^3 n) time
###################################################################
def Perm_to_Benes(perm):
  """整数の配列で表された置換を log n 個のBenesネットワークに変換
     O(n log^3 n) time

  Args:
      perm (整数配列): 置換の配列表現

  Returns:
      N ([Benes]): 置換のBenesネットワーク表現
  """
  n = len(perm)
  w = blog(n-1)+1
  N = [None] * w
  perm_new = perm.copy()
  for k in range(w):
    print("Perm to Benes step:", k)
    b = [None] * n
    for i in range(n):
      b[i] = perm_new[i][k] # k ビット目を取り出す
    net, _ = Benes(b)
    N[k] = net
    perm_new = Benes_apply(perm_new, net, False) # O(n log^2 n) time
  return N

###################################################################
### wビット整数の配列 X をソートする w 個のBenesネットワークを求める
### O(w^2n log n) time
###################################################################
def Benes_sort(X, w):
  """wビット整数の配列 X をソートする w 個のBenesネットワークを求める
     O(w^2n log n) time

  Args:
      X (整数配列): ソートする配列
      w (整数(平文)): X のビット数

  Returns:
      N ([Benes]): X をソートする w 個のBenesネットワークの配列
      X_new (整数配列): ソートされた配列
  """
  n = len(X)
  N = [None] * w
  X_new = X.copy()
  for k in range(w):
    #print("Benes_sort step:", k)
    b = [None] * n
    for i in range(n):
      b[i] = X_new[i][k]
    net, _ = Benes(b)
    N[k] = net
    X_new = Benes_apply(X_new, net, False) # O(wn log n) time
  return N, X_new

###################################################################
### k 個のBenesネットワークで表された置換をw-bit整数の配列 x に適用
### (置換の場合は k = log n)
### inverse == True の場合は逆置換を計算
### O(kwn log n) time
###################################################################
def AppPerm(x, N, inverse = False):
  """k 個のBenesネットワークで表された置換をw-bit整数の配列 x に適用
     (置換の場合は k = log n)
     O(kwn log n) time
  Args:
      x (整数配列): 置換を適用する配列
      N ([Benes]): k 個のBenesネットワークで表された置換
      inverse (bool, optional): True の場合は逆置換を計算

  Returns:
      整数配列: Nで表された置換を x に適用した配列
  """
  n = len(x)
  #k = blog(n-1)+1
  k = len(N)
  print("AppPerm: x =", type(x[0]))
  if type(x[0]) == tfhe.PyLweInt:
    t = 1
  elif type(x[0]) == tfhe.PyLweShort:
    t = 2
#    x = [v.to_packed() for v in x]
  else:
    print("AppPerm: unsupported type:", type(x[0]))
  #x = [v.to_packed() for v in x]
  if inverse:
    for i in reversed(range(k)):
      print("AppPerm step:", i)
      x = Benes_apply(x, N[i], inverse)
  else:
    for i in range(k):
      print("AppPerm step:", i)
      x = Benes_apply(x, N[i], inverse)
#  if t == 2:
#    x = [v.to_short() for v in x]
  #else:
  #  x = [v.to_short() for v in x]
  return x

###################################################################
### Benesネットワークで表された置換の逆置換を表すBenesネットワークを返す
###################################################################
def Benes_inverse(N):
  """Benesネットワークで表された置換の逆置換を表すBenesネットワークを返す

  Args:
      N (Benes): Benesネットワーク

  Returns:
      N_inv (Benes): 逆置換のBenesネットワーク
  """
  if N == []:
    return []
  sl, netU, netL, sr = N
  if netU == [] and netL == [] and sr == []:
    return [sl, [], [], []] # n == 2 の場合，SL のみで表される
  return [sr, Benes_inverse(netU), Benes_inverse(netL), sl]

###################################################################
### k 個のBenesネットワークで表された置換の逆置換を表すBenesネットワークの配列を返す
###################################################################
def Perm_inverse(N):
  """k 個のBenesネットワークで表された置換の逆置換を表すBenesネットワークの配列を返す

  Args:
      N ([Benes]): Benesネットワークの配列

  Returns:
      N_inv ([Benes]): 逆置換のBenesネットワークの配列
  """
  k = len(N)
  N_inv = [None] * k
  for i in range(k):
    N_inv[i] = Benes_inverse(N[k-1-i])
  return N_inv


###################################################################
### k 個のBenesネットワークで表された置換を整数の配列に変換 (k <= log n)
### inverse == True の場合は逆置換を計算
### O(kn log^2 n) time
###################################################################
def Benes_to_Perm(N, inverse = False):
  """k 個のBenesネットワークで表された置換を整数の配列に変換
     k <= log n

  Args:
      N ([Benes]): Benesネットワークの配列
      inverse (bool, optional): True の場合は逆置換を計算

  Returns:
      pi (整数配列): Nで表された置換の配列表現
  """
  n = len(N[0][0])*2
  w = blog(n-1)+1
  id = [encrypt_int(i, w) for i in range(n)]
  pi = AppPerm(id, N, inverse)
  return pi

###################################################################
### 配列で表された置換と (log n 個の) Benesネットワークで表された置換を合成した置換を表す配列を計算
### (Benesネットワークに変換するには Perm_to_Benes を使う)
### O(n log^3 n) time
###################################################################
def Compose(pi, N, inverse = False):
  sigma = AppPerm(pi, N, inverse)
  return sigma

###############################################################
### 各位置で，それより左の接頭辞での or を計算
### （先頭は必ず 0 になる）
###############################################################
#def AllPrefixOr(X, inclusive=False):
#  n = len(X)
#  O = [None] * n
#  r = LWE_FALSE
#  for i in range(n):
#    if inclusive:
#      r = r | X[i]
#    O[i] = r
#    if not inclusive:
#      r = r | X[i]
#  return O

###############################################################
### 各位置で，それより右の接尾辞での or を計算
### （先頭は必ず 0 になる）
###############################################################
#def AllSuffixOr(X, inclusive=False):
#  n = len(X)
#  O = [None] * n
#  r = LWE_FALSE
#  for i in range(n-1, -1, -1):
#    if inclusive:
#      r = r | X[i]
#    O[i] = r
#    if not inclusive:
#      r = r | X[i]
#  return O

################################################################
### w ビット整数 x の unit vector 表現を返す
### O(n) 時間
################################################################
#def unitv(x):
#    w = x.get_bit_size()
#    n = 1 << w
#
#    x_bits = x.to_bit()
#
#    F = [LWE_TRUE]
#    for d in range(w):
#        s = 1 << d
#        F_next = [None] * (s*2)
#        b = x_bits[w-1-d]
#        for i in range(s):
#            F_next[2*i] = F[i] & ~b
#            F_next[2*i+1] = F[i] & b
#        F = F_next
#    return F

################################################################
### w bit配列 V を整数 k だけ左回転
### O(wn log n) 時間
################################################################
#def rotate_left(V, k):
#    n = len(V)
#    u = unitv(k)
#    o = AllSuffixOr(u)
#    pi, _ = Benes(o)
#    V_rot = Benes_apply(V, pi, False)
#    return V_rot

################################################################
### w bit配列 V を整数 k だけ右回転
### O(wn log n) 時間
################################################################
#def rotate_right(V, k):
#    n = len(V)
#    u = unitv(k)
#    o = AllSuffixOr(u)
#    pi, _ = Benes(o)
#    V_rot = Benes_apply(V, pi, True)
#    return V_rot

def rlex(x, d):
  if d == 0:
    return x

  if (x >> d) & 1: # 下半分
    x = x ^ 1
  if x & 1 == 1:
    return rlex(x // 2, d - 1) + (1 << d)
  else:
    return rlex(x // 2, d - 1)

def inv_butterfly2(X, pos, length):
  Y = [None] * length
  for i in range(0,length//2):
    Y[i*2] = X[pos + i]
    Y[i*2+1] = X[pos + i+length//2]
  for i in range(length):
    X[pos + i] = Y[i]

#################################################################
### スイッチを再帰的な表現に変換
#################################################################
def Perm_to_Benes_1bit_sub(S):
  k = len(S)
  if k == 1:
    return [S[0], [], [], []]
  n = len(S[0])
  SL = S[0]
  U = [None] * (k-1)
  L = [None] * (k-1)
  SR = [encrypt(0)] * (n//2) + [encrypt(1)] * (n//2)
  for i in range(k-1):
    U[i] = S[i+1][0:n//2]
    L[i] = S[i+1][n//2:n]
  NU = Perm_to_Benes_1bit_sub(U)
  NL = Perm_to_Benes_1bit_sub(L)

  N = [SL, NU, NL, SR]
  return N

#################################################################
### 0,1 列の安定ソートを表す置換の配列表現から Benes ネットワークを求める
### 0 と 1 の数は同じとする
### 配列の長さは2冪とする
### 入力 X: 置換を表す配列
### 出力 N: 置換 X の Benes ネットワーク表現
### O(n log^2 n) 時間
#################################################################
def Perm_to_Benes_1bit(X):
  """0,1 列の安定ソートを表す置換の配列表現から Benes ネットワークを求める
     0 と 1 の数は同じとする
     配列の長さは2冪とする
     O(n log^2 n) 時間

  Args:
      X (整数配列): 置換を表す配列

  Raises:
      ValueError: X の長さが2冪でない場合

  Returns:
      N (Benes): 置換 X の Benes ネットワーク表現
  """
  n0 = len(X)
  k = blog(n0-1)+1
  m = 1 << k
  n = n0
  if n < m:
  #  raise ValueError("Perm_to_Benes_1bit: length of X must be a power of 2")
    tmp = []
    for i in range(m-n):
      tmp.append(encrypt_int(n+i, k))
    X = X + tmp
  Y = [None] * m
  for i in range(m):
    Y[rlex(i, k-1)] = X[i]
  #print("Y:", [yi.dec() for yi in Y])

  S = [None] * k

  for d in range(k-1, -1, -1):
    S[d] = []
    for i in range(m//2):
      y = Y[2*i]
      #c = (y >> d) & 1
      c = y[d]
      S[d].append(c)
      Y[2*i], Y[2*i+1] = ApplySwitch(c, Y[2*i], Y[2*i+1])
    if d == 0:
      break
    pos = 0
    for g in range(1<<d-1):
      inv_butterfly2(Y, pos, (1<<k-d+1))
      pos += (1 << k-d+1)

  N = Perm_to_Benes_1bit_sub(S)
  return Benes_inverse(N)
#  return N

def Perm_to_Str_1bit(P):
  """0,1 列 X の安定ソートを表す置換の配列表現 P から X を求める
     0 と 1 の数は同じとする
     配列の長さは2冪とする
     O(n log n) 時間

  Args:
      P (整数配列): 置換を表す配列

  Returns:
      X (0,1列): 置換 P に対応する 0,1 列
  >>> P = enc_array_int([0, 4, 1, 2, 5, 6, 3, 7], 3)
  >>> X = Perm_to_Str_1bit(P)
  >>> dec_array(X)
  [0, 1, 0, 0, 1, 1, 0, 1]
  >>> N, _ = Benes(X)
  >>> Q = Benes_to_Perm([N])
  >>> dec_array(Q)
  [0, 2, 3, 6, 1, 4, 5, 7]
  """ 
  n = len(P)
  k = blog(n-1)+1
  Q = [encrypt_int(i, k) for i in range(n)]
  X = [None] * n
  for i in range(n):
    if i < n//2:
      #lt = Q[i] < P[i]
      #c = lt
      c = Q[i] < P[i]
    else:
      #lt = Q[i] < P[i]
      #eq = Q[i] == P[i]
      #c = lt | eq
      c = Q[i] <= P[i]
    X[i] = c
  return X

####################################################################
def convert_vecbenes_to_Benes(N):
#  print("convert_vecbenes_to_Benes:")
#  print("N len:", N.get_len())
#  print("network:", [v.dec() for v in N.get_v()])
  network = N.get_v()
  nlen = N.get_len()

  sl_len = nlen[0]
  if sl_len == 1:
#    print("switch_c: ")
#    for l in network:
#        print("{} ", l.dec())
#    print("\n")
    return [network, [], [], []]
  sr_len = nlen[1]
  if sr_len == 1:
#    print("switch_c: ")
#    for l in network:
#        print("{} ", l.dec())
#    print("\n")
    return [network, [], [], []]

  n1_len = nlen[2]
  n2_len = nlen[3]
  n1_llen = nlen[4]
  n2_llen = nlen[5]
  sl_pos = 0
  sr_pos = sl_pos + sl_len
  n1_pos = sr_pos + sr_len
  n2_pos = n1_pos + n1_len

  sl = network[sl_pos : (sl_pos + sl_len)]
  sr = network[sr_pos : (sr_pos + sr_len)]
#  print("sl_len:", sl_len, " sr_len:", sr_len, " n1_len:", n1_len, " n2_len:", n2_len)
#  print("sl: ")
#  for l in sl:
#      print("{} ", l.dec())
#  print("\n")
#  print("sr: ")
#  for l in sr:
#      print("{} ", l.dec())
#  print("\n")
  network1 = []
  if n1_len > 0:
    #print("network1: ")
    network1 = tfhe.PyVecBenes()
    network1.set(network[n1_pos:(n1_pos + n1_len)], nlen[6 : 6+n1_llen])
    #network1.v = network[n1_pos:(n1_pos + n1_len)]
    #network1.len = nlen[4:(4+n1_len)]
    network1 = convert_vecbenes_to_Benes(network1)
    #print("\n")

  network2 = []
  if n2_len > 0:
    #print("network2: ")
    network2 = tfhe.PyVecBenes()
    network2.set(network[n2_pos:(n2_pos + n2_len)], nlen[6+n1_llen : 6+n1_llen+n2_llen])
    #network2.v = network[n2_pos:(n2_pos + n2_len)]
    #network2.len = nlen[4+n1_len:(4+n1_len+n2_len)]
    network2 = convert_vecbenes_to_Benes(network2)
    #print("\n")

  return [sl, network1, network2, sr]


####################################################################

flatten = lambda x: [z for y in x for z in (flatten(y) if hasattr(y, '__iter__') and not isinstance(y, str) else (y,))]
flatten_dec = lambda x: [z.dec() for y in x for z in (flatten(y) if hasattr(y, '__iter__') and not isinstance(y, str) else (y,))]


if __name__ == '__main__':
#  import doctest
#  from csclib_tfhe.pycsclib_tfhe import enc_array_int, enc_array, dec_array
#  doctest.testmod()

#  k = 3
#  X = [0, 2, 3, 6, 1, 4, 5, 7]
#  X_enc = [encrypt_int(x, k) for x in X]
#  print("X:", [xi.dec() for xi in X_enc])
#  N = Perm_to_Benes_1bit(X_enc)
#  Y = [0, 1, 2, 3, 4, 5, 6, 7]
#  Y_enc = [encrypt_int(x, k) for x in Y]
#  Y_new = Benes_apply(Y_enc, N)
#  print("Y_new:", [xi.dec() for xi in Y_new])

  from csclib_tfhe.pycsclib_tfhe import enc_array_int, enc_array, dec_array
  P = enc_array_int([0, 4, 1, 2, 5, 6, 3, 7], 3)
  X = Perm_to_Str_1bit(P)
  print(dec_array(X))
  N, _ = Benes(X)
  Q = Benes_to_Perm([N])
  print(dec_array(Q))
  Q2 = Benes_to_Perm([N], inverse=True)
  print(dec_array(Q2))
  


from csclib_tfhe import *
from csclib_tfhe.benes import Benes, Benes_apply, AppPerm
from csclib_tfhe.gencycle import Propagate
#from csclib_tfhe.pycsclib_tfhe import blog
from csclib_tfhe import LWE_TRUE, LWE_FALSE

def vneg(idx):
  N = len(idx)
  idx_neg = [None] * N
  for i in range(N):
    idx_neg[i] = LWE_TRUE - idx[i]
  return idx_neg

##########################################################################
### 配列の要素をまとめて取り出す． 
# v はアクセスしたい配列．長さ U
# idx は v のアクセスしたい要素の添え字の配列を 1 進数表現にしたもの（添え字は単調増加）．
# idx の値の最大値は len(v) 未満 
# idx 中の 0 の数は v の値域の大きさと等しい
# idx 中の 1 の数はアクセス回数
# 1進数表現とは，0^k 1 で整数 k を表す方法
##########################################################################
def BatchAccessUnary(v, idx):
  """配列の要素をまとめて取り出す

  Args:
      v (整数配列): アクセスしたい配列．長さ U
      idx (0,1列): v のアクセスしたい要素の添え字の配列を 1 進数表現にしたもの（添え字は単調増加）．
                    idx の値の最大値は len(v) 未満
                    idx 中の 0 の数は v の値域の大きさと等しい
                    idx 中の 1 の数はアクセス回数
                    1進数表現とは 0^k 1 で整数 k を表す方法

  Returns:
      W (整数配列): 取り出した要素の配列. 長さ |idx|-|v|
  >>> dec_array(BatchAccessUnary(enc_array_int([0,1,2,3,4,5,6,7], 3), enc_array([0,0,1,0,0,1,1,1,0,0,0,1,0])))
  [1, 3, 3, 3, 6]
  """
  U = len(v)
  N = len(idx)
  sigma, _ = Benes(idx)
  X = v + ([tfhe.PyLweInt(0, v[0].get_bit_size())] * (N-U))
  Y = AppPerm(X, [sigma], inverse=True)
  Z = Propagate(vneg(idx), Y)
  W = AppPerm(Z, [sigma], inverse=False)
  return W[U:]

if __name__ == '__main__':
#
#  X = [0,0,1,0,0,1,1,1,0,0,1,0,1,0,1,1]
#  n = len(X)
#  k = blog(n-1)+1
#  X_enc = [encrypt(x) for x in X]
#
#  V = [tfhe.PyLweInt(i, k) for i in range(8)]
#
#  W = BatchAccessUnary(V, X_enc)
#  print("W:", [wi.dec() for wi in W])
  import doctest
  from csclib_tfhe.pycsclib_tfhe import enc_array_int, enc_array, dec_array
  doctest.testmod()



from csclib_tfhe import _csclib_tfhe as tfhe


def pylwe_xor_int(self, other):
  if type(other) == int: other = tfhe.PyLweInt(other, self.get_bit_size())
  return self.xor(other)
def pylwe_mod_int(self, other):
  if type(other) == int: other = tfhe.PyLweInt(other, self.get_bit_size())
  return self._mod(other)
def pylwe_eq_int(self, other):
  if type(other) == int: other = tfhe.PyLweInt(other, self.get_bit_size())
  return self.eq(other)
def pylwe_neq_int(self, other):
  if type(other) == int: other = tfhe.PyLweInt(other, self.get_bit_size())
  return self.__neq__(other)
def pylwe_lt_int(self, other):
  if type(other) == int: other = tfhe.PyLweInt(other, self.get_bit_size())
  return self.lt(other)
def pylwe_gt_int(self, other):
  if type(other) == int: other = tfhe.PyLweInt(other, self.get_bit_size())
  return self.gt(other)
def pylwe_add_int(self, other):
  if type(other) == int:
    if other < 0: other += (1 << self.get_bit_size())
    other = tfhe.PyLweInt(other, self.get_bit_size())
  if type(self) == int:
    if self < 0: self += (1 << other.get_bit_size())
    self = tfhe.PyLweInt(self, other.get_bit_size())
  return self.add(other)
def pylwe_sub_int(self, other):
  if type(other) == int:
    if other < 0: other += (1 << self.get_bit_size())
    other = tfhe.PyLweInt(other, self.get_bit_size())
  return self.sub(other)
def pylwe_array_int(x, k):
  return [tfhe.PyLweInt(xi, k) for xi in x]

def pylwe_mul_int(self, other):
  #if type(other) == int:
  #  if other < 0: other += (1 << self.get_bit_size())
  #  other = tfhe.PyLweInt(other, self.get_bit_size())
  #if type(self) == int:
  #  if self < 0: self += (1 << other.get_bit_size())
  #  self = tfhe.PyLweInt(self, other.get_bit_size())
  return self.mul(other)


tfhe.PyLweInt.__xor__ = pylwe_xor_int
tfhe.PyLweInt.__rxor__ = pylwe_xor_int
tfhe.PyLweInt.__mod__ = pylwe_mod_int
tfhe.PyLweInt.__eq__ = pylwe_eq_int
tfhe.PyLweInt.__neq__ = pylwe_neq_int
tfhe.PyLweInt.__lt__ = pylwe_lt_int
tfhe.PyLweInt.__gt__ = pylwe_gt_int
tfhe.PyLweInt.__add__ = pylwe_add_int
tfhe.PyLweInt.__radd__ = pylwe_add_int
tfhe.PyLweInt.__sub__ = pylwe_sub_int
tfhe.PyLweInt.array = pylwe_array_int

def pylwe_xor_short(self, other):
  if type(other) == int: other = tfhe.PyLweShort(other)
  return self.xor(other)
def pylwe_mod_short(self, other):
  if type(other) == int: other = tfhe.PyLweShort(other)
  return self._mod(other)
def pylwe_eq_short(self, other):
  if type(other) == int: other = tfhe.PyLweShort(other)
  return self.eq(other)
def pylwe_neq_short(self, other):
  if type(other) == int: other = tfhe.PyLweShort(other)
  return self.__neq__(other)
def pylwe_lt_short(self, other):
  if type(other) == int: other = tfhe.PyLweShort(other)
  return self.lt(other)
def pylwe_gt_short(self, other):
  if type(other) == int: other = tfhe.PyLweShort(other)
  return self.gt(other)
def pylwe_add_short(self, other):
  if type(other) == int:
    if other < 0: other += (1 << 2) # 2 が埋め込まれているので注意
    other = tfhe.PyLweShort(other)
  if type(self) == int:
    if self < 0: self += (1 << 2)
    self = tfhe.PyLweShort(self, other)
  return self.add(other)
def pylwe_sub_short(self, other):
  if type(other) == int:
    if other < 0: other += (1 << 2)
    other = tfhe.PyLweShort(other)
  return self.sub(other)
def pylwe_array_short(x):
  return [tfhe.PyLweShort(xi) for xi in x]
tfhe.PyLweShort.__xor__ = pylwe_xor_short
tfhe.PyLweShort.__rxor__ = pylwe_xor_short
tfhe.PyLweShort.__mod__ = pylwe_mod_short
tfhe.PyLweShort.__eq__ = pylwe_eq_short
tfhe.PyLweShort.__neq__ = pylwe_neq_short
tfhe.PyLweShort.__lt__ = pylwe_lt_short
tfhe.PyLweShort.__gt__ = pylwe_gt_short
tfhe.PyLweShort.__add__ = pylwe_add_short
tfhe.PyLweShort.__radd__ = pylwe_add_short
tfhe.PyLweShort.__sub__ = pylwe_sub_short
tfhe.PyLweShort.array = pylwe_array_short

def pylwe_slice(self, range_obj):
  #print("slice:", type(range_obj))
  if type(range_obj) == slice:
    #print(range_obj.start, range_obj.stop, range_obj.step)
    start, stop, step = range_obj.start, range_obj.stop, range_obj.step
    if range_obj.start is None:
      start = 0
    if range_obj.stop is None:
      stop = self.get_bit_size()
    if range_obj.step is None:
      step = 1
    return self.slice(start, stop, step)
  else:
    return self.getitem(range_obj)
tfhe.PyLweShortArray.__getitem__ = pylwe_slice
tfhe.PyLweIntArray.__getitem__ = pylwe_slice
tfhe.PyLweInt.__getitem__ = pylwe_slice

def blog(x):
  l = -1
  while x > 0:
    x >>= 1
    l += 1
  return l

def enc(x):
  return tfhe.PyLweShort(x)

def enc_int(x, bit):
  return tfhe.PyLweInt(x, bit)

def enc_packed(x, bit):
  return tfhe.PyLwePacked(x, bit)

def enc_array(arr):
    return [enc(val) for val in arr]

def enc_array_int(arr, bit):
    return [enc_int(val, bit) for val in arr]

def enc_array_packed(arr, bit):
    return [enc_packed(val, bit) for val in arr]

def dec_array(arr):
    return [val.dec() for val in arr]

def pylargelut_lookup(self, x):
  if type(x) == tfhe.PyLweInt:
    x = self.calibrate(x)
  return self.lookup(x)
tfhe.PyLargeLUT.__getitem__ = pylargelut_lookup

tfhe.PyLweShortArray.__len__ = lambda self: self.len()
tfhe.PyLweIntArray.__len__ = lambda self: self.len()

def pylwe_array_iter(self):
  for i in range(self.len()):
    yield self.getitem(i)

tfhe.PyLweIntArray.__iter__ = pylwe_array_iter
tfhe.PyLweShortArray.__iter__ = pylwe_array_iter

def pylweintarray_concat(self, other):
  if type(self) == int:
    self = tfhe.PyLweIntArray([self], other.get_bit_size())
  elif type(self) == list:
    if type(self[0]) == tfhe.PyLweInt:
      self = tfhe.PyLweIntArray(self)
    elif type(self[0]) == int:
      self = tfhe.PyLweIntArray(self, other.get_bit_size())

  if type(other) == tfhe.PyLweIntArray:
    return self.concat(other)
  if type(other) == list:
    if type(other[0]) == tfhe.PyLweInt:
      other = tfhe.PyLweIntArray(other)
      return self.concat(other)
    elif type(other[0]) == int:
      r = tfhe.PyLweIntArray(other, self.get_bit_size())
      return self.concat(r)
  raise TypeError("Unsupported type for concat:", type(other))

tfhe.PyLweIntArray.__matmul__ = pylweintarray_concat
tfhe.PyLweIntArray.__rmatmul__ = pylweintarray_concat


def pylweshortarray_concat(self, other):
  if type(self) == int:
    self = tfhe.PyLweShortArray([self], other.get_bit_size())
  elif type(self) == list:
    if type(self[0]) == tfhe.PyLweShort:
      #self = tfhe.PyLweShortArray.from_list(self)
      self = tfhe.PyLweShortArray(self)
    elif type(self[0]) == int:
      self = tfhe.PyLweShortArray(self, other.get_bit_size())

  if type(other) == tfhe.PyLweShortArray:
    return self.concat(other)
  if type(other) == list:
    if type(other[0]) == tfhe.PyLweShort:
      other = tfhe.PyLweShortArray(other)
      return self.concat(other)
    elif type(other[0]) == int:
      r = tfhe.PyLweShortArray(other, self.get_bit_size())
      return self.concat(r)
  raise TypeError("Unsupported type for concat:", type(other))

def pylweshortarray_mul(self, other):
  if type(other) != int:
    raise TypeError("Unsupported type for mul:", type(other))
  l = []
  for _ in range(other):
    l += self.v
  return tfhe.PyLweShortArray(l)


tfhe.PyLweShortArray.__matmul__ = pylweshortarray_concat
tfhe.PyLweShortArray.__rmatmul__ = pylweshortarray_concat

#tfhe.PyLweShortArray.__mul__ = pylweshortarray_mul
